import numpy as np


def reward_function(params):
    # 입력 파라미터 추출
    all_wheels_on_track = params['all_wheels_on_track']
    waypoints = params['waypoints']
    closest_waypoints = params['closest_waypoints']
    heading = params['heading']
    steering_angle = params['steering_angle']
    steps = params['steps']
    distance_from_center = params['distance_from_center']
    track_width = params['track_width']
    progress = params['progress']
    speed = params.get('speed', 0.0)
    prev_speed = params.get('prev_speed', 0.0)
    is_clockwise = params.get('is_clockwise', True)

    # 속도 및 코너링 임계값 정의
    SPEED_THRESHOLD = 3.0
    MAX_SPEED_THRESHOLD = 4.0
    CORNERING_THRESHOLD = 20
    REWARD_PROGRESS_THRESHOLD = 100

    # 기본 보상 초기화
    reward = 1.0

    # 1. 트랙 이탈 확인
    if not all_wheels_on_track:
        return 1e-3  # 트랙을 벗어나면 즉시 큰 패널티 적용

    # 2. 경로 유지 보상
    next_waypoint = waypoints[closest_waypoints[1]]
    prev_waypoint = waypoints[closest_waypoints[0]]
    track_direction = np.arctan2(next_waypoint[1] - prev_waypoint[1], next_waypoint[0] - prev_waypoint[0])
    track_direction = np.degrees(track_direction)
    direction_diff = abs(track_direction - heading)

    if direction_diff < 10.0:
        reward += 1.0  # 경로 유지 시 보상
    elif direction_diff < 20.0:
        reward += 0.5  # 약간의 경로 벗어남
    else:
        reward += 0.2  # 경로와 크게 벗어나면 보상이 줄어듦

    # 3. 코너링 강도와 방향 계산
    prev_track_direction = np.arctan2(prev_waypoint[1] - waypoints[max(closest_waypoints[0] - 1, 0)][1],
                                      prev_waypoint[0] - waypoints[max(closest_waypoints[0] - 1, 0)][0])
    prev_track_direction = np.degrees(prev_track_direction)
    direction_change = track_direction - prev_track_direction  # 방향 변화량 계산

    # 코너의 방향을 판단하여 좌/우 코너 결정
    if direction_change > 0:  # 오른쪽 코너
        is_right_corner = True
    else:  # 왼쪽 코너
        is_right_corner = False
    direction_change = abs(direction_change)

    # 4. 코너링 구간에서의 속도 및 벽면 보상
    if direction_change > CORNERING_THRESHOLD:
        if SPEED_THRESHOLD < speed < MAX_SPEED_THRESHOLD:
            reward += 1.0  # 코너링 구간에서 적절한 속도를 유지하면 보상
        else:
            reward -= 0.5  # 과속 또는 속도 감소 부족 시 패널티

        # 예상 코너링 방향에 따라 미리 반대편 벽면으로 이동
        if is_right_corner:  # 오른쪽 코너 예상
            if distance_from_center < 0.5 * track_width:
                reward += 0.5  # 오른쪽 코너 전에 왼쪽으로 미리 이동 시 보상
        else:  # 왼쪽 코너 예상
            if distance_from_center > 0.5 * track_width:
                reward += 0.5  # 왼쪽 코너 전에 오른쪽으로 미리 이동 시 보상

        # 벽면 이동 보상 (코너 구간에서만 적용)
        if is_clockwise:
            if distance_from_center > 0.5 * track_width:
                reward += 1.0  # 시계 방향 코너에서 오른쪽 벽면으로 이동 시 보상
            else:
                reward -= 0.5  # 이동 실패 시 패널티
        else:
            if distance_from_center < 0.5 * track_width:
                reward += 1.0  # 반시계 방향 코너에서 왼쪽 벽면으로 이동 시 보상
            else:
                reward -= 0.5  # 이동 실패 시 패널티

    # 5. 직선 구간에서의 속도 보상 및 중앙 유지
    if direction_change < CORNERING_THRESHOLD:
        if speed > SPEED_THRESHOLD:
            reward += 1.2  # 직선 구간에서 빠른 속도 유지 시 보상
        if distance_from_center <= 0.1 * track_width:
            reward += 1.5  # 직선 구간에서 중앙을 유지하면 추가 보상

    # 6. 트랙 중앙에서 벗어난 경우 속도를 줄이면 보상
    if distance_from_center > 0.5 * track_width and speed < SPEED_THRESHOLD:
        reward += 0.5  # 트랙 중앙에서 벗어난 경우 속도를 줄이면 보상

    # 7. 스티어링 각도에 따른 보상
    if abs(steering_angle) > 15:
        reward *= 0.8  # 큰 스티어링 각도는 보상 감소 (지그재그 방지)

    # 8. 진행률 기반 속도 최적화
    if progress > 50:
        if direction_change < CORNERING_THRESHOLD and speed > SPEED_THRESHOLD:
            reward += 1.0  # 직선 구간에서 적절한 속도 유지 시 보상
        elif direction_change > CORNERING_THRESHOLD and speed < SPEED_THRESHOLD:
            reward += 0.5  # 코너링 구간에서 적절히 속도를 줄이면 보상
        else:
            reward -= 0.5  # 속도 유지 실패 시 패널티

    # 9. 마지막 구간에서 속도 보상 강화
    if progress > 90 and speed > SPEED_THRESHOLD:
        reward += 2.0  # 마지막 10% 구간에서 빠르게 주행 시 보상 강화

    # 10. 부드러운 주행 보상 (급격한 가속/감속 방지)
    if abs(speed - prev_speed) < 0.5:
        reward += 1.0  # 부드러운 속도 변화 시 보상

    return float(reward)
